/************************************************************************/
/*									*/
/*  Procedures to read an ispell affix file and dictionary files and	*/
/*  to build a finite automaton that recognises the same words.		*/
/*									*/
/************************************************************************/

# ifndef IND_AFFIX_H
# define IND_AFFIX_H

#   include	<stdio.h>
#   include	<string.h>
#   include	<iconv.h>
#   include	<ind.h>
#   include	<charnames.h>
#   include	<utilIndexSet.h>
#   include	<utilIndexMapping.h>
#   include	<appDebugon.h>

/************************************************************************/
/*									*/
/*  An affix rule, For simplicity a fixed maximum size for the strings	*/
/*  is assumed.								*/
/*									*/
/************************************************************************/

# define	AFFMAXLENGTH	16

/*
 * A string in a replacement rule: The strings are encoded in UTF-16
 */
typedef struct RuleString
    {
    int			rsLength;
    unsigned short	rsUtf16[AFFMAXLENGTH+1];
    } RuleString;

typedef struct ApplicabilityMask
    {
    int		amNegative;
    IndexSet	amSet;
    } ApplicabilityMask;

typedef struct AffixRule
    {
    int			arWhatSide;
    int			arShift;
    int			arCross;
    int			arMaskLength;
    struct AffixRule *	arNext;
    ApplicabilityMask	arMasks[AFFMAXLENGTH];

			/*
			 * The replacement rules
			 */
    RuleString		arReplaceFrom;
    RuleString		arReplaceWith;
    int			arSuffixNumber;
    } AffixRule;


# define	AFFstrchrMAXLENGTH	3

/*
 * Map a byte string in the affix file to an UTF-16 unicode value.
 */
typedef struct StringChar
    {
			/*
			 * The byte sequence and its length
			 */
    unsigned char	scString[AFFstrchrMAXLENGTH+1];
    int			scLength;
			/*
			 * The unicode meaning
			 */
    unsigned short	scMeaningUtf16;
    } StringChar;

typedef struct AffixRules
    {
    StringChar *	arStringChars;
    int			arStringCharCount;
    IndexSet		arWordChars;
    IndexSet		arBoundaryChars;
			/*
			 * An array of linked lists of rules.
			 */
    AffixRule *		arsRules[256];
    } AffixRules;


typedef struct AffixThrough
    {
    const char *	atEncodingIn;
    const char *	atEncodingOut;
    iconv_t		atIconv;
    void *		atVoidInd;
    AffixRules *	atRules;
    void *		atSuffixInd;
    IndexMapping	atSuffixMapping;
    } AffixThrough;

extern AffixRules * indAffixReadFile(	const char *		filename,
					const AffixThrough *	at );

extern int indAffixApplyRules(	const AffixRules *	ars,
				int			prefixOnly,
				const unsigned short *	word,
				int			len,
				const char *		rules,
				void *			voidind );

extern void indAffixListRules(	const void *		voidars );

extern void affixInitThrough(	AffixThrough *	at );

extern void affixCleanThrough(	AffixThrough *	at );

extern int affixToUtf16(	iconv_t *		iconv,
				unsigned short *	scratch,
				int			scratchSize,
				const char *		word,
				int			wordLength );

extern void affixLogUtf16( const unsigned short * w );

extern void indAffixListRule(	int			rule,
				const AffixRule *	ar );

extern void indAffixListMask(	const ApplicabilityMask *	am );

extern void affixInitRule(	AffixRule *	ar );
extern void affixFreeRule(	AffixRule *	ar );
extern void affixInitRules(	AffixRules *	ars );
extern void affixFreeRules(	AffixRules *	ars );

extern int affixChar(		const AffixRules *	ars,
				const char *		buf,
				iconv_t			icv,
				int			lower,
				int *			pPos );

extern int affixCollectSuffixes(	AffixThrough *		at );

extern int affixApplyPrefixRules(	const AffixRules *	ars,
					int			cross,
					const unsigned short *	word,
					int			len,
					const char *		rules,
					void *			ind );

# endif /* IND_AFFIX_H */
